Program TowersOfHanoi;
{ ͻ }
{                           TOWERS OF HANOI                          }
{                         by Sven Neumann 1996                       }
{ Ķ }
{                    E-MAIL: CSITC@Mail.link.net.mt                  }
{ Ķ }
{  This is an old "Game" or better a mathematical challange.         }
{  A tower of disks (various sizes) has to be moved from the left    }
{  of three poles to the right one, while applying following rules:  }
{                                                                    }
{           - Only one disk can be moved at a time                   }
{           - A disk can only be put onto a larger one               }
{                                                                    }
{  For each additional disk the number of turns needed to complete   }
{  the disk shift doubles. So for a tower of 64 disks a computer     }
{  able to do 100 turns per second (A pentium is MUCH slower!)       }
{  would take about 4000 years to complete.                          }
{ Ķ }
{  Hanoi is an example often used to demonstrate recursion in        }
{  A-Level books. The common version usually just writes lines of    }
{  Text informing about the movements of disks.                      }
{  This version Displays the disks and poles graphically(Text-mode)  }
{  and the disk movements can be displayed animated!                 }
{ ͼ }

Uses Crt; { Crt is used for some of the Key handling routines (Keypressed) }

Type TowersType = array[1..3,0..20] of integer;
      {              Three towers with space for 20 Disks)            }
      { This limit is necessary because otherwise the towers wouldn't }
      {                   fit on the screen properly.                 }

Var Towers:TowersType;                { A variable of Type TowersType }
    Count,Height,maxturn,TimeDelay:longint;
    Animation:Boolean;

Procedure Disc(x,y,size:integer);
{ This Procedure draws a disk at position (X,Y) of size: SIZE       }
{ It is mainly used for the animation and screen updates            }
{ Disks are drawn at the largest size possible, without overlapping }
{ Numbers indicate the "real" size of the disks                     }
  var c1,c2,DrawSize:integer;
  begin
    DrawSize:=Size;
    if DrawSize>9 then DrawSize:=10;
    TextColor(15);
    GotoXY(x-10 ,26-y);
    If Size>=0 then TextColor(size mod 16);
    if (size mod 16)=0 then TextColor(6);{ No black Disks allowed }
    case Drawsize of
      -2:begin
        GotoXY(x-10 ,26-y);
        Write('                     ');
      end;
      -1:begin
        TextColor(8);
        GotoXY(x-10 ,26-y);
        Write('                    ');
      end;
      0:begin
        GotoXY(x,26-y);Write('');
      end;
      1..maxint:begin
        GotoXY(x-Drawsize+0,26-y);Write('');            { These characters are }
        GotoXY(x-Drawsize+1,26-y);Write('');            { used for the shading }
        GotoXY(x-Drawsize+Drawsize*2-1,26-y);Write(''); { of the disk edges.   }
        GotoXY(x-Drawsize+Drawsize*2,26-y);Write('');   { (3D effect, crude..) }
        for C1:=2 to Drawsize*2-2 do begin
          GotoXY(x-Drawsize+c1,26-y);Write('');end;
      end;
    end;
    if Size>-1 then begin
      TextBackground(size mod 16);
      if size mod 16=0 then TextBackground(6);
      if size mod 16=8 then TextBackground(7);
      TextColor(0);
      GotoXY(x,26-y);
      Write(Size+1);
      TextBackground(0);
    end;
  end;

Procedure EndProg(Status:Byte);     { Handle Exit from program }
  var Count:byte;
  begin
    if Status=0 then begin      { Normal exit, towers finished }
      TextColor(2);
      GotoXY(20,10);
      Write(' Finished! Press [ENTER] to exit ...');
      repeat until readkey=chr(13);
      TextColor(15);         { Reset Color }
      ClrScr;
      asm                         { Switch on cursor  }
        mov ah,1                  { Function 1        }
        mov cx,$0607              { Parameter - COLOR MONITOR ! }
        int $10                   { Interrupt 10h     }
      end;                        { end ASM           }
      halt(Status);
    end else begin           { User abort }
      TextColor(4);
      GotoXY(10,10);
      Write(' Aborted! Press any key to exit or [C] to continue ...');
      repeat until keypressed;
      if upcase(readkey)<>'C' then begin  { Confimation }
        TextColor(15);     { Reset Color }
        ClrScr;
        halt(Status);
      end else begin   { No exit, so redraw part of the screen }
        GotoXY(10,10);
        Write('                                                      ');
        if Height>15 then begin
          if Towers[1,0]>15 then Disc(20,16,Towers[1,3])
          else Disc(20,16,-1);
          if Towers[2,0]>15 then Disc(40,16,Towers[2,3])
          else Disc(40,16,-1);
          if Towers[3,0]>15 then Disc(60,16,Towers[3,3])
          else Disc(60,16,-1);
        end;
      end;
    end;
  end;

Procedure CheckKey;  { Check for keypress, and empty key buffer }
  begin;
    while keypressed do begin
      case readkey of
        'A','a':Animation:=not Animation;
        '+':if TimeDelay>1 then TimeDelay:=TimeDelay-1;
        '-':inc(TimeDelay);
        chr(27):EndProg(1);
      end;
    end;
    if (MaxTurn-Count<7)then begin { Almost finished, so slow down }
      if TimeDelay<20 then TimeDelay:=20;
      Animation:=true;
    end;
  end;

Procedure lift(pole:integer);
{ Lift the top most disk of Tower: pole to two lines above the top. }
  var x1,c1,c2:integer;
  begin
    x1:=Towers[pole,0]+1;
    for c1:=x1 to Height do begin
      Disc(pole*20,c1+1,Height-Towers[pole,x1]);
      Disc(pole*20,c1,-1);
      delay(TimeDelay);
      CheckKey; { Check after each step, to avoid buffer overflow }
    end;
    for c1:=Height+1 to Height+1 do begin
      Disc(pole*20,c1+1,Height-Towers[pole,x1]);
      Disc(pole*20,c1,-2);
      delay(TimeDelay);
      CheckKey; { Check after each step, to avoid buffer overflow }
    end;
  end;

Procedure drop(pole,size:integer);
{ Drop a disk to the top most position of tower: pole. }
  var x1,c1,c2:integer;
  begin
    x1:=Towers[pole,0]+1;
    for c1:=Height+1 downto Height do begin
      Disc(pole*20,c1,size);
      Disc(pole*20,c1+1,-2);
      delay(TimeDelay);
      CheckKey; { Check after each step, to avoid buffer overflow }
    end;
    for c1:=height-1 downto x1-1 do begin
      Disc(pole*20,c1,size);
      Disc(pole*20,c1+1,-1);
      delay(TimeDelay);
      CheckKey; { Check after each step, to avoid buffer overflow }
    end;
  end;

Procedure shift(source,dest,size:integer);
{ Shift a disk from source to dest }
  var x1,c1,c2,step:integer;
  begin
    if source<dest then step:=1 else step:=-1;
    x1:=source*20;
    repeat
      Disc(x1-step,Height+2,-2);
      Disc(x1,Height+2,size);
      delay(TimeDelay div 3*2);
      x1:=x1+step;
      CheckKey; { Check after each step, to avoid buffer overflow }
    until x1=dest*20+1;
  end;

Procedure PrintStatus;
{ Print some status information }
  begin
    GotoXY(30,3);
    TextColor(15);
    write(Count,' of ',maxturn);
  end;

Procedure Move(source,dest:integer);
{ Move a disk from source to dest with or without animation }
  var c1,c2:integer;
  begin
    Towers[dest,Towers[dest,0]+1]:=Towers[source,Towers[source,0]]; { Disc at new position  }
    Towers[dest,0]:=Towers[dest,0]+1;                         { Delete disc at old position }
    Towers[source,0]:=Towers[source,0]-1;                     { Correct height              }
    if Animation then begin  { Lift, shift and drop the disc }
      lift(source);
      shift(source,dest,Height-Towers[dest,Towers[dest,0]]);
      drop(dest,Height-Towers[dest,Towers[dest,0]]);
    end else begin
      Disc(dest*20,Towers[dest,0],Height-Towers[dest,Towers[dest,0]]);
      { No animation, so slow down a bit }
      Delay(TimeDelay*2);
      CheckKey;
    end;
    inc(Count);
    PrintStatus;
  end;

Procedure Transfer(N,source,dest,tmp:integer);
{ Main Procedure. This calls itself to recursively move smaller }
{ stacks of disks, until only one disk has to be moved.         }
  begin
    if N=1 then
      Move(source,dest)
    else begin
      transfer(N-1,source,tmp,dest);
      move(source,dest);
      transfer(N-1,tmp,dest,source);
    end;
  end;

begin
  clrscr;
  TextColor(15);             { Here I could have implemented save input! }
  Write('Enter Height: ');   { Try what happens when you enter a string! }
  readln(Height);{ Try to improve it! hint: convert a string to a number }
  if Height>20 then begin { Higher then 20 does not fit on screen properly }
    Height:=20;
    Writeln('Towers too high! 20 is maximum...');
    delay(500);               { wait to show message }
  end else if Height<1 then begin { smaller then 1 causes errors }
    Height:=1;
    Writeln('Towers too small! 1 is minimum...');
    delay(500);               { wait to show message }
  end;
  asm                         { Switch off cursor }
    mov ah,1                  { Function 1        }
    mov cx,$2020              { Parameter in CX   }
    int $10                   { Interrupt 10h     }
  end;                        { end ASM           }
  ClrScr;
  TimeDelay:=30;              { Default Delay: 30 }
  Animation:=true;            { Animation: ON     }
  MaxTurn:=2;                 { Calculate Number of Turns       }
  for count:=1 to Height-1 do { Somewhat awkward, but it        }
    MaxTurn:=MaxTurn*2;       { works. Binary Number!(2^Height) }
  Maxturn:=MaxTurn-1;
  GotoXY(19,1);
  TextColor(14);Write('_');   { Some creds and graphics...HEHE! }
  TextColor(4);Write('');
  TextColor(6);Write(' ');
  TextColor(13);
  Write('Towers of Hanoi by Sven Neumann 1996');
  TextColor(6);Write(' ');
  TextColor(4);Write('');  { ||   || }
  TextColor(14);Write('_'); {\||/ \||/}
  TextColor(7);GotoXY(7,2); { \/   \/ "Read the friendly Manual"(RTFM)-Stuff }
  Write('([ESC] to stop, [+]/[-] to in/decrease speed, [A] to toggle Animation)');
  TextColor(15);GotoXY(25,3);
  write('Move:');
  Towers[1,0]:=Height;
  Towers[2,0]:=0;
  Towers[3,0]:=0;
  for Count:=1 to Height do begin  { Initialize towers }
    Towers[1,Count]:=Count;
    Disc(20,Count,Height-Towers[1,Count]);
    Disc(40,Count,-1);                         { Poles }
    Disc(60,Count,-1);
  end;
  Count:=0;
  PrintStatus;
  Transfer(Height,1,3,2);     { This is actually the whole thing... }
  EndProg(0);                 { Terminate normally! (Or user break) }
end.
